@@ -85,6 +85,10 @@ read_zarr_array <- function(zarr_array_path, index, s3_client) {
       version_to = 3
     )
   }
+  metadata$configured_codecs <- .configure_codecs(
+    metadata$codecs,
+    operation = "decode"
+  )
 
   ## if no index provided we will return everything
   if (missing(index)) {
@@ -149,14 +153,16 @@ read_zarr_array <- function(zarr_array_path, index, s3_client) {
   warn <- chunk$warning[1]
   chunk_data <- chunk$chunk_data
 
-  ## extract the required elements from the chunk
-  selection <- R.utils::extract(
-    chunk_data,
-    indices = index_in_chunk,
-    drop = FALSE
-  )
+  if (!is.null(chunk_data)) {
+    ## extract the required elements from the chunk
+    chunk_data <- R.utils::extract(
+      chunk_data,
+      indices = index_in_chunk,
+      drop = FALSE
+    )
+  }
 
-  return(list(selection, index_in_result, warning = warn))
+  return(list(chunk_data, index_in_result, warning = warn))
 }
 
 
@@ -201,20 +207,21 @@ read_data <- function(
       )
     }
   )
-
   ## predefine our array to be populated from the read chunks
   output <- array(metadata$fill_value, dim = lengths(index))
 
   ## proceed in serial and update the output with each chunk selection in turn
   for (i in seq_along(chunk_selections)) {
-    index_in_result <- chunk_selections[[i]][[2]]
-    cmd <- .create_replace_call(
-      x_name = "output",
-      idx_name = "index_in_result",
-      idx_length = length(index_in_result),
-      y_name = "chunk_selections[[i]][[1]]"
-    )
-    eval(parse(text = cmd))
+    if (!is.null(chunk_selections[[i]][[1]])) {
+      index_in_result <- chunk_selections[[i]][[2]]
+      cmd <- .create_replace_call(
+        x_name = "output",
+        idx_name = "index_in_result",
+        idx_length = length(index_in_result),
+        y_name = "chunk_selections[[i]][[1]]"
+      )
+      eval(parse(text = cmd))
+    }
     warn <- max(warn, chunk_selections[[i]]$warning[1])
   }
   return(list(output = output, warn = warn))
@@ -263,6 +270,9 @@ get_decompressed_chunk_size <- function(datatype, dimensions) {
 #'   `metadata`, but when dealing with edge chunks, which may overlap the true
 #'   extent of the array the returned array should be smaller than the chunk
 #'   shape.
+#' @param fill Logical of length 1.  If `TRUE`, missing chunks will be filled
+#'    with the fill value from the array metadata.  If `FALSE` (the default),
+#'    missing chunks will return `NULL`.
 #'
 #' @returns A list of length 2.  The entries should be names "chunk_data" and
 #'   "warning". The first is an array containing the decompressed chunk values,
@@ -275,7 +285,8 @@ read_chunk <- function(
   chunk_name,
   metadata,
   s3_client = NULL,
-  alt_chunk_dim = NULL
+  alt_chunk_dim = NULL,
+  fill = FALSE
 ) {
   chunk_file <- paste0(zarr_array_path, chunk_name)
 
@@ -284,8 +295,8 @@ read_chunk <- function(
   }
 
   if (is.null(s3_client)) {
-    size <- file.size(chunk_file)
     if (file.exists(chunk_file)) {
+      size <- file.size(chunk_file)
       compressed_chunk <- readBin(con = chunk_file, what = "raw", n = size)
     } else {
       compressed_chunk <- NULL
@@ -303,37 +314,47 @@ read_chunk <- function(
     }
   }
 
-  ## either decompress and format the chunk data
-  ## or create a new chunk based on the fill value
-  if (!is.null(compressed_chunk)) {
-    decompressed_chunk <- .decompress_chunk(compressed_chunk, metadata)
-
-    bytes_codec_config <- metadata$codecs[["bytes"]]$configuration %||%
-      NA_character_
-    decompressed_chunk <- codec_endian_decode(
-      decompressed_chunk,
-      bytes_codec_config,
-      ifelse(
-        # For unicode, nbytes actually is sizeof(int) * nchar
-        metadata$datatype$base_type == "unicode",
-        4L,
-        metadata$datatype$nbytes
+  # Missing chunks are filled with the fill value in the whole array.
+  # We generally don't need to re-fill here and this saves resources.
+  # A notable exception when we need to fill is in update_zarr_array().
+  if (is.null(compressed_chunk)) {
+    if (fill) {
+      return(
+        list(
+          "chunk_data" = array(
+            metadata$fill_value,
+            dim = unlist(metadata$chunk_grid$configuration$chunk_shape)
+          ),
+          "warning" = 0L
+        )
       )
+    } else {
+      return(
+        list(
+          "chunk_data" = NULL,
+          "warning" = 0L
+        )
+      )
+    }
+  }
+
+  # Bytes -> Bytes codecs
+  for (codec in metadata$configured_codecs[["bytes_bytes"]]) {
+    compressed_chunk <- codec(
+      bytes = compressed_chunk
     )
-    converted_chunk <- .format_chunk(
-      decompressed_chunk,
-      metadata,
-      alt_chunk_dim
-    )
-    # FIXME: run array -> array codecs here
-  } else {
-    converted_chunk <- list(
-      "chunk_data" = array(
-        metadata$fill_value,
-        dim = unlist(metadata$chunk_grid$configuration$chunk_shape)
-      ),
-      "warning" = 0L
-    )
+  }
+  decompressed_chunk <- compressed_chunk
+
+  # Bytes -> Array codecs
+  converted_chunk <- .format_chunk(
+    decompressed_chunk,
+    metadata,
+    alt_chunk_dim
+  )
+  # Array -> Array codecs
+  for (codec in metadata$configured_codecs[["array_array"]]) {
+    converted_chunk[[1]] <- do.call(codec, list(converted_chunk[[1]]))
   }
 
   return(converted_chunk)
@@ -369,11 +390,25 @@ read_chunk <- function(
 .format_chunk <- function(decompressed_chunk, metadata, alt_chunk_dim) {
   datatype <- metadata$datatype
 
+  bytes_codec_config <- metadata$codecs[["bytes"]]$configuration %||%
+    NA_character_
+
+  decompressed_chunk <- codec_bytes_decode(
+    decompressed_chunk,
+    bytes_codec_config,
+    ifelse(
+      # For unicode, nbytes actually is sizeof(int) * nchar
+      metadata$datatype$base_type == "unicode",
+      4L,
+      metadata$datatype$nbytes
+    )
+  )
+
   ## It doesn't seem clear if the on disk chunk will contain the overflow
   ## values or not, so we try both approaches.
   actual_chunk_size <- length(decompressed_chunk) / datatype$nbytes
   if (
-    !is.null(metadata$codecs[["vlen-utf8"]]) ||
+    !is.null(metadata$codecs[["vlen_utf8"]]) ||
       (actual_chunk_size ==
         prod(unlist(metadata$chunk_grid$configuration$chunk_shape)))
   ) {
@@ -383,7 +418,7 @@ read_chunk <- function(
   }
 
   if (datatype$base_type == "string") {
-    if (!is.null(metadata$codecs[["vlen-utf8"]])) {
+    if (!is.null(metadata$codecs[["vlen_utf8"]])) {
       converted_chunk <- .format_object(decompressed_chunk, metadata, datatype)
       dim(converted_chunk[[1]]) <- chunk_dim
     } else {
@@ -412,92 +447,6 @@ read_chunk <- function(
     )
   }
 
-  if (!is.null(metadata$codecs$transpose)) {
-    transpose_config <- metadata$codecs[["transpose"]]$configuration
-    # FIXME: R is already F ordered, so we reverse the order in config
-    converted_chunk[[1]] <- codec_transpose_encode(
-      converted_chunk[[1]],
-      rev(unlist(transpose_config$order)) + 1L
-    )
-  }
-
   names(converted_chunk) <- c("chunk_data", "warning")
   return(converted_chunk)
 }
-
-#' Decompress a chunk in memory
-#'
-#' R has internal decompression tools for zlib, bz2 and lzma compression.  We
-#' use external libraries bundled with the package for blosc and lz4
-#' decompression.
-#'
-#' @param compressed_chunk Raw vector holding the compressed bytes for this
-#'   chunk.
-#' @param metadata List produced by `.read_array_metadata()` with the contents of
-#'   the `.zarray` file.
-#'
-#' @returns An array with the number of dimensions specified in the Zarr
-#'   metadata.  In most cases it will have the same size as the Zarr chunk,
-#'   however in the case of edge chunks, which overlap the extent of the array,
-#'   the returned chunk will be smaller.
-#'
-#' @importFrom utils tail
-#' @keywords internal
-.decompress_chunk <- function(compressed_chunk, metadata) {
-  decompressor <- intersect(
-    names(metadata$codecs),
-    c("blosc", "zlib", "gzip", "bz2", "lzma", "lz4", "zstd")
-  )
-  datatype <- metadata$datatype
-  buffer_size <- get_decompressed_chunk_size(
-    datatype,
-    dimensions = metadata$chunk_grid$configuration$chunk_shape
-  )
-
-  if (length(decompressor) == 0) {
-    decompressed_chunk <- compressed_chunk
-  } else if (decompressor == "blosc") {
-    decompressed_chunk <- .Call(
-      "decompress_chunk_BLOSC",
-      compressed_chunk,
-      PACKAGE = "Rarr"
-    )
-  } else if (decompressor %in% c("zlib", "gzip")) {
-    decompressed_chunk <- memDecompress(
-      from = compressed_chunk,
-      type = "gzip",
-      asChar = FALSE
-    )
-  } else if (decompressor == "bz2") {
-    decompressed_chunk <- memDecompress(
-      from = compressed_chunk,
-      type = "bzip2",
-      asChar = FALSE
-    )
-  } else if (decompressor == "lzma") {
-    decompressed_chunk <- memDecompress(
-      from = compressed_chunk,
-      type = "xz",
-      asChar = FALSE
-    )
-  } else if (decompressor == "lz4") {
-    ## numpy codecs stores the original size of the buffer in the first 4 bytes
-    decompressed_chunk <- .Call(
-      "decompress_chunk_LZ4",
-      tail(x = compressed_chunk, n = -4L),
-      buffer_size,
-      PACKAGE = "Rarr"
-    )
-  } else if (decompressor == "zstd") {
-    decompressed_chunk <- .Call(
-      "decompress_chunk_ZSTD",
-      compressed_chunk,
-      buffer_size,
-      PACKAGE = "Rarr"
-    )
-  } else {
-    stop("Unsupported compression tool")
-  }
-
-  return(decompressed_chunk)
-}
