@@ -305,7 +305,7 @@ zarr_overview <- function(zarr_array_path, s3_client, as_data_frame = FALSE) {
     # 'name' MUST be a plain string.
     names(metadata$codecs) <- vapply(
       metadata$codecs,
-      function(x) x$name,
+      function(x) gsub("-", "_", x$name, fixed = TRUE),
       character(1)
     )
     if (length(metadata$shape) == 0) {
@@ -379,6 +379,8 @@ zarr_overview <- function(zarr_array_path, s3_client, as_data_frame = FALSE) {
       "int" = as.integer(val),
       "uint" = as.integer(val),
       "complex" = as.complex(val),
+      "string" = as.character(val),
+      "unicode" = as.character(val),
       val
     )
   }
