@@ -9,20 +9,31 @@
   }
 
   ## if data type was supplied directly, always use that
-  if (!data_type %in% c("|i1", "<i2", "<i4", "<f4", "<f8", "|S", "<U", "|b1")) {
+  supported_types <- c(
+    "|i1",
+    "<i2",
+    "<i4",
+    "<i8",
+    "|u1",
+    "<u2",
+    "<u4",
+    "<u8",
+    "<f4",
+    "<f8",
+    "|S",
+    "<U",
+    "|b1"
+  )
+  if (data_type %notin% supported_types) {
     data_type <- switch(
       data_type,
       "integer" = "<i4",
       "double" = "<f8",
       "character" = "|S",
       "logical" = "|b1",
-      NULL
-    )
-  }
-
-  if (is.null(data_type)) {
-    stop(
-      "Currently only able to write integer, double, character and logical arrays"
+      stop(
+        "Currently only able to write integer, double, character and logical arrays"
+      )
     )
   }
 
@@ -33,6 +44,11 @@
       "|i1" = 0L,
       "<i2" = 0L,
       "<i4" = 0L,
+      "<i8" = 0L,
+      "|u1" = 0L,
+      "<u2" = 0L,
+      "<u4" = 0L,
+      "<u8" = 0L,
       "<f4" = 0,
       "<f8" = 0,
       "|S" = "",
@@ -282,9 +298,7 @@ write_zarr_array <- function(
   .compress_and_write_chunk(
     input_chunk = chunk_in_mem,
     chunk_path = chunk_path,
-    metadata,
-    compressor = metadata$compressor,
-    data_type_size = .parse_datatype(metadata$dtype)$nbytes
+    metadata
   )
 
   return(invisible(TRUE))
@@ -335,26 +349,23 @@ update_zarr_array <- function(zarr_array_path, x, index) {
   )
   index <- check_index(index, metadata = metadata)
 
-  data_type <- switch(
-    storage.mode(x),
-    "integer" = "<i",
-    "double" = "<f",
-    "character" = c("|S", "<U", ">U"),
-    "logical" = "|b",
+  existing_storage <- switch(
+    metadata_v3$datatype$base_type,
+    "uint" = "integer",
+    "int" = "integer",
+    "float" = "double",
+    "bool" = "logical",
+    "string" = "character",
+    "unicode" = "character",
     NULL
   )
-  if (!substr(metadata$dtype, 1, 2) %in% data_type) {
+  if (!identical(storage.mode(x), existing_storage)) {
     stop("New data is not of the same type as the existing array.")
   }
 
   zarr_dim <- unlist(metadata$shape)
   chunk_dim <- unlist(metadata$chunks)
 
-  ## convert strings to Unicode if required
-  if (grepl("<U|>U", x = metadata$dtype, fixed = FALSE)) {
-    x <- .unicode_to_int(input = x, typestr = metadata$dtype)
-  }
-
   ## coerce x to the same shape as the zarr to be updated
   x <- array(x, dim = lengths(index))
 
@@ -430,7 +441,8 @@ update_zarr_array <- function(zarr_array_path, x, index) {
   chunk_in_mem <- read_chunk(
     zarr_array_path = path,
     chunk_name = chunk_id,
-    metadata = metadata_v3
+    metadata = metadata_v3,
+    fill = TRUE
   )[["chunk_data"]]
 
   ## extract the new values from x and insert them into the chunk
@@ -442,14 +454,11 @@ update_zarr_array <- function(zarr_array_path, x, index) {
     "y"
   )
   eval(parse(text = cmd))
-
   ## re-compress updated chunk and write back to disk
   .compress_and_write_chunk(
     input_chunk = chunk_in_mem,
     chunk_path = chunk_path,
     metadata,
-    compressor = metadata$compressor,
-    data_type_size = metadata$datatype$nbytes,
     is_base64 = (metadata$datatype$base_type == "unicode")
   )
 }
@@ -460,12 +469,6 @@ update_zarr_array <- function(zarr_array_path, x, index) {
 #'   converted to a raw vector before compression.
 #' @param chunk_path Character string giving the path to the chunk that should
 #'   be written.
-#' @param compressor A "compressor" function that returns a list giving the
-#'   details of the compression tool to apply.  See [compressors] for more
-#'   details.
-#' @param data_type_size An integer giving the size of the original datatype.
-#'   This is passed to the blosc algorithm, which seems to need it to achieve
-#'   any compression.
 #' @param is_base64 When dealing with Py_unicode strings we convert them to
 #' base64 strings for storage in our intermediate R arrays.  This argument
 #' indicates if base64 is in use, because the conversion to raw in .as_raw
@@ -479,40 +482,50 @@ update_zarr_array <- function(zarr_array_path, x, index) {
   input_chunk,
   chunk_path,
   metadata,
-  compressor = use_zlib(),
-  data_type_size,
   is_base64 = FALSE
 ) {
+  # Array to array codecs
   if (metadata$order == "C") {
-    input_chunk <- codec_transpose_decode(
+    input_chunk <- codec_transpose_encode(
       input_chunk,
       indices = rev(seq_along(dim(input_chunk)))
     )
   }
 
+  # Array to bytes codecs
+  ## convert strings to Unicode if required
+  if (metadata$datatype$base_type == "unicode") {
+    input_chunk <- .unicode_to_int(
+      input = input_chunk,
+      typestr = metadata$dtype
+    )
+  }
+
   ## the compression tools need a raw vector
   raw_chunk <- .as_raw(
     as.vector(input_chunk),
-    nchar = data_type_size,
+    nchar = metadata$datatype$nbytes,
     is_base64 = is_base64
   )
 
   # Endianness in unicode is handled during the conversion to/from
   if (metadata$datatype$base_type != "unicode") {
-    raw_chunk <- codec_endian_encode(
+    raw_chunk <- codec_bytes_encode(
       raw_chunk,
       endian = metadata$datatype$endian,
       metadata$datatype$nbytes
     )
   }
 
+  # Bytes to bytes codecs
+  compressor <- metadata$compressor
   if (is.null(compressor)) {
     compressed_chunk <- raw_chunk
   } else if (compressor$id == "blosc") {
     compressed_chunk <- .Call(
       "compress_chunk_BLOSC",
       raw_chunk,
-      as.integer(data_type_size),
+      as.integer(metadata$datatype$nbytes),
       PACKAGE = "Rarr"
     )
   } else if (compressor$id == "zlib") {
