@@ -44,3 +44,24 @@ test_that("v2 and v3 return identical results", {
 
   expect_identical(u8_v2, u8_v3)
 })
+
+test_that("uint8 zarr array can be written", {
+  uint8_zarr <- withr::local_tempfile(fileext = ".zarr")
+
+  content_uint8_array <- array(1:24, dim = c(4, 3, 2))
+
+  write_zarr_array(
+    content_uint8_array,
+    uint8_zarr,
+    data_type = "|u1",
+    chunk_dim = c(2, 2, 1),
+    compressor = NULL
+  )
+
+  roundtrip_uint8_array <- read_zarr_array(uint8_zarr)
+
+  expect_identical(
+    content_uint8_array,
+    roundtrip_uint8_array
+  )
+})
